package de.hft.stuttgart.citydoctor2.gui;

import java.util.ArrayList;
import java.util.List;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.ErrorVisitor;
import de.hft.stuttgart.citydoctor2.check.error.AllPolygonsWrongOrientationError;
import de.hft.stuttgart.citydoctor2.check.error.AttributeInvalidError;
import de.hft.stuttgart.citydoctor2.check.error.AttributeMissingError;
import de.hft.stuttgart.citydoctor2.check.error.AttributeValueWrongError;
import de.hft.stuttgart.citydoctor2.check.error.ConsecutivePointSameError;
import de.hft.stuttgart.citydoctor2.check.error.DependenciesNotMetError;
import de.hft.stuttgart.citydoctor2.check.error.MultipleConnectedComponentsError;
import de.hft.stuttgart.citydoctor2.check.error.NestedRingError;
import de.hft.stuttgart.citydoctor2.check.error.NonManifoldEdgeError;
import de.hft.stuttgart.citydoctor2.check.error.NonManifoldVertexError;
import de.hft.stuttgart.citydoctor2.check.error.NonPlanarPolygonDistancePlaneError;
import de.hft.stuttgart.citydoctor2.check.error.NonPlanarPolygonNormalsDeviation;
import de.hft.stuttgart.citydoctor2.check.error.NotCeilingError;
import de.hft.stuttgart.citydoctor2.check.error.NotFloorError;
import de.hft.stuttgart.citydoctor2.check.error.NotGroundError;
import de.hft.stuttgart.citydoctor2.check.error.NotWallError;
import de.hft.stuttgart.citydoctor2.check.error.NullAreaError;
import de.hft.stuttgart.citydoctor2.check.error.PointTouchesEdgeError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonHoleOutsideError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonInteriorDisconnectedError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonIntersectingRingsError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonSameOrientationError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonWithoutSurfaceError;
import de.hft.stuttgart.citydoctor2.check.error.PolygonWrongOrientationError;
import de.hft.stuttgart.citydoctor2.check.error.RingDuplicatePointError;
import de.hft.stuttgart.citydoctor2.check.error.RingEdgeIntersectionError;
import de.hft.stuttgart.citydoctor2.check.error.RingNotClosedError;
import de.hft.stuttgart.citydoctor2.check.error.RingTooFewPointsError;
import de.hft.stuttgart.citydoctor2.check.error.SchematronError;
import de.hft.stuttgart.citydoctor2.check.error.SolidNotClosedError;
import de.hft.stuttgart.citydoctor2.check.error.SolidSelfIntError;
import de.hft.stuttgart.citydoctor2.check.error.SurfaceUnfragmentedError;
import de.hft.stuttgart.citydoctor2.check.error.DegeneratedRingError;
import de.hft.stuttgart.citydoctor2.check.error.TooFewPolygonsError;
import de.hft.stuttgart.citydoctor2.check.error.UnknownCheckError;
import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import javafx.scene.paint.Color;

public class ListErrorVisitor implements ErrorVisitor {
	
	private HighlightController controller;
	private TriangulatedGeometry geom;
	
	public ListErrorVisitor(HighlightController controller) {
		this.controller = controller;
	}
	
	public void setGeometry(TriangulatedGeometry geom) {
		this.geom = geom;
	}

	@Override
	public void visit(PolygonHoleOutsideError err) {
		List<LinearRing> highlightedRings = new ArrayList<>();
		highlightedRings.add(err.getPolygon().getExteriorRing());
		highlightedRings.addAll(err.getHolesOutside());
		controller.highlight(highlightedRings, geom);
	}

	@Override
	public void visit(NonManifoldEdgeError err) {
		controller.highlightEdges(err.getEdges(), geom);
	}

	@Override
	public void visit(MultipleConnectedComponentsError err) {
		controller.highlightPolygons(err.getComponents(), geom);
	}

	@Override
	public void visit(NestedRingError err) {
		List<LinearRing> highlightedRings = new ArrayList<>();
		highlightedRings.add(err.getPolygon().getExteriorRing());
		highlightedRings.add(err.getInnerRing());
		controller.highlight(highlightedRings, geom);
	}

	@Override
	public void visit(NonManifoldVertexError err) {
		controller.highlightPolygons(err.getComponents(), geom);
		controller.addHighlight(err.getVertex(), geom, Color.BLUEVIOLET);
	}

	@Override
	public void visit(PolygonWrongOrientationError err) {
		controller.highlightEdges(err.getEdges(), geom);
	}

	@Override
	public void visit(PolygonSameOrientationError err) {
		List<LinearRing> highlightedRings = new ArrayList<>();
		highlightedRings.add(err.getPolygon().getExteriorRing());
		highlightedRings.add(err.getInnerRing());
		controller.highlight(highlightedRings, geom);
	}

	@Override
	public void visit(SolidNotClosedError err) {
		controller.highlightEdges(err.getErrorEdges(), geom);
	}

	@Override
	public void visit(DependenciesNotMetError err) {
		// don't display
	}

	@Override
	public void visit(UnknownCheckError err) {
		// don't display
	}

	@Override
	public void visit(RingNotClosedError err) {
		controller.highlight(err.getRing(), geom);
	}

	@Override
	public void visit(ConsecutivePointSameError err) {
		controller.highlight(err.getRing(), geom);
		controller.addHighlight(err.getVertex1(), geom, Color.BLACK);
		controller.addHighlight(err.getVertex2(), geom, Color.BLACK);
	}

	@Override
	public void visit(AllPolygonsWrongOrientationError err) {
		// don't display
	}

	@Override
	public void visit(PolygonInteriorDisconnectedError err) {
		controller.highlight(err.getConnectedRings(), geom);
	}

	@Override
	public void visit(NullAreaError err) {
		// don't display
	}

	@Override
	public void visit(RingTooFewPointsError err) {
		controller.highlight(err.getRing(), geom);
	}

	@Override
	public void visit(NonPlanarPolygonNormalsDeviation err) {
		controller.highlight(err.getPolygon(), geom);
	}

	@Override
	public void visit(NonPlanarPolygonDistancePlaneError err) {
		controller.highlight(err.getPolygon(), geom);
		controller.addHighlight(err.getVertex(), geom, Color.BLACK);
	}

	@Override
	public void visit(PolygonIntersectingRingsError err) {
		List<LinearRing> rings = new ArrayList<>();
		rings.add(err.getIntersectingRings().getValue0());
		rings.add(err.getIntersectingRings().getValue1());
		controller.highlight(rings, geom);
	}

	@Override
	public void visit(SolidSelfIntError err) {
		// nothing to display for now
	}

	@Override
	public void visit(TooFewPolygonsError err) {
		// don't display
	}

	@Override
	public void visit(RingDuplicatePointError err) {
		controller.highlight(err.getRing(), geom);
		controller.addHighlight(err.getVertex1(), geom, Color.BLACK);
		controller.addHighlight(err.getVertex2(), geom, Color.BLACK);
	}

	@Override
	public void visit(RingEdgeIntersectionError err) {
		List<Edge> list = new ArrayList<>();
		list.add(err.getEdge1());
		list.add(err.getEdge2());
		controller.highlightEdges(list, geom);
		controller.addHighlight(new Vertex(err.getIntersection()), geom, Color.BLACK);
	}

	@Override
	public void visit(PointTouchesEdgeError err) {
		controller.highlight(err.getEdge(), geom);
		controller.addHighlight(err.getVertex(), geom, Color.BLACK);
	}

	@Override
	public void visit(CheckError err) {
		// nothing to display
	}

	@Override
	public void visit(NotCeilingError err) {
		// nothing to display
	}

	@Override
	public void visit(NotFloorError err) {
		// nothing to display
	}

	@Override
	public void visit(NotWallError err) {
		// nothing to display
	}

	@Override
	public void visit(NotGroundError err) {
		// nothing to display
	}

	@Override
	public void visit(SchematronError err) {
		// nothing to display
	}

	@Override
	public void visit(SurfaceUnfragmentedError err) {
		// nothing to display
	}

	@Override
	public void visit(DegeneratedRingError err) {
		controller.highlight(err.getRing(), geom);
	}

	@Override
	public void visit(AttributeMissingError err) {
		// nothing to display
	}

	@Override
	public void visit(AttributeValueWrongError err) {
		// nothing to display
	}

	@Override
	public void visit(AttributeInvalidError err) {
		// nothing to display
	}

	@Override
	public void visit(PolygonWithoutSurfaceError err) {
		controller.highlight(err.getPolygon(), geom);
	}
}
